{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}

module Data.Number.Instances where

import Data.Number.Types
import Data.Number.Internal

instance Functor Continued where
  fmap _ E       = E
  fmap f (M x)   = M (fmap f x)
  fmap f (x:|xs) = f x :| fmap f xs 

instance Num Number where
  (+) = operator (0, 1,  1, 0, 1, 0, 0, 0)
  (-) = operator (0, 1, -1, 0, 1, 0, 0, 0)
  (*) = operator (0, 0,  0, 1, 1, 0, 0, 0)

  abs (M x) = x
  abs x = x

  negate (M x) = x
  negate x = M x

  fromInteger = toNumber . fromIntegral

  signum E     = 0
  signum (M _) = -1
  signum _     = 1

instance Real Number where
  toRational = fromNumber

instance Fractional Number where
  (/) = operator (0, 1, 0, 0, 0, 0, 1, 0)
  fromRational = toNumber


-- Helpers --

fromNumber :: RealFrac a => Number -> a
fromNumber E         = 0
fromNumber (M x)     = negate (fromNumber x)
fromNumber (x :| E)  = fromIntegral x
fromNumber (x :| xs) = fromIntegral x + 1 / (fromNumber xs)

toNumber :: (Show a, RealFrac a) => a -> Number
toNumber 0  = E
toNumber x0
  | x0 < 0    = M (n :| toNumber x1)
  | otherwise =    n :| toNumber x1
  where
    (n,f) = properFraction (abs x0)
    x1 | f < 1e-6  = 0
       | otherwise = 1/f


-- constants --

φ :: Number
φ = 1 :| φ

σ :: Number
σ = σ' 0 where
  σ' n = n :| σ' (succ n)

π :: Number
π = toNumber pi

e :: Number
e = fmap a σ where 
  a n | p == 0    = 2*q
      | otherwise = 1
    where (q, p) = quotRem n 3
